# -*- coding: utf-8 -*-

import json
import struct
import threading
import logging
from logging.handlers import RotatingFileHandler
import serial
import time
import os
import sys
import xml.etree.ElementTree as ET


from modbus_tk import modbus_rtu
from modbus_tk import LOGGER, utils

from define import DeviceOTABase, UpdateCode, OTAError, Log, AgentBasic, OTAResult, ENABLE_DEBUG

WORK_DIR = os.path.dirname(os.path.realpath(sys.argv[0]))

PROGRAM_VERSION = '1.0.0'
PROGRAM_NAME = 'Amitime-OTA'


def initial_logger(filepath='update.log', enable_stdout=False, enable_debug=False):
    Log.logger = LOGGER
    log_format = logging.Formatter('%(asctime)s [%(threadName)s][%(levelname)s]: %(message)s')
    rt_handler = RotatingFileHandler(filepath, mode='a', maxBytes=8 * 1024, backupCount=1)
    rt_handler.setFormatter(log_format)
    Log.logger.addHandler(rt_handler)
    if enable_stdout:
        rf_handler = logging.StreamHandler(sys.stdout)
        rf_handler.setFormatter(log_format)
        Log.logger.addHandler(rf_handler)
    if enable_debug:
        Log.logger.setLevel(logging.DEBUG)
    else:
        Log.logger.setLevel(logging.INFO)


class OTAFile(object):
    PACKAGE_BYTE_SIZE = 128

    FILE_TYPE_BIN = 01
    FILE_TYPE_HEX = 02

    def __init__(self, filedir):
        self.__filedir = filedir
        self.__filepath = ''
        self.__filesize = 0
        self.__filetype = 0
        self.__filestream = None

        self.__package_num = 0
        self.__package_index = 0
        self.__package_left_bytes = 0

    def check_self(self):
        if not os.path.exists(self.__filedir):
            raise OTAError(info="待升级文件所在目录不存在")
        for file_name in os.listdir(self.__filedir):
            basename, extension = os.path.splitext(file_name)
            file_path = os.path.join(self.__filedir, file_name)
            if os.path.isfile(file_path) \
                    and extension.upper() in ['.HEX', '.BIN']:
                self.__filepath = file_path
                self.__filesize = os.path.getsize(self.__filepath)
                Log.info('获取升级文件（%d bytes）：%s' % (self.__filesize, file_name))
                self.__package_num = self.__filesize // self.PACKAGE_BYTE_SIZE
                self.__package_left_bytes = self.__filesize % self.PACKAGE_BYTE_SIZE
                if self.__package_left_bytes > 0:
                    self.__package_num += 1
                else:
                    self.__package_left_bytes = self.PACKAGE_BYTE_SIZE
                if extension.upper() == 'HEX':
                    self.__filetype = self.FILE_TYPE_HEX
                else:
                    self.__filetype = self.FILE_TYPE_BIN
                return True
        raise OTAError(info="未找到合适的升级文件")

    def _open_filestream(self):
        if self.__filestream:
            return
        if os.path.isfile(self.filepath) and self.__filesize > 0:
            self.__filestream = open(self.__filepath, 'rb')

    def _close_filestream(self):
        if self.__filestream:
            self.__filestream.close()
            self.__filestream = None

    def get_next_package(self):
        self._open_filestream()
        self.__package_index += 1
        eof = (self.__package_index == self.__package_num)
        length = 0
        if eof:
            length = self.__package_left_bytes
            data = self.__filestream.read(length)
            self._close_filestream()
        else:
            length = self.PACKAGE_BYTE_SIZE
            data = self.__filestream.read(length)
        return eof, self.__package_index, length, data

    def set_next_package_index(self, index):
        if index > self.__package_num:
            return False
        self._open_filestream()
        self.__package_index = index - 1
        self.__filestream.seek(self.__package_index * self.PACKAGE_BYTE_SIZE, 0)
        return True

    @property
    def filepath(self):
        return self.__filepath

    @property
    def filename(self):
        return os.path.basename(self.__filepath)

    @property
    def filesize(self):
        return self.__filesize

    @property
    def filetype(self):
        return self.__filetype

    @property
    def package_num(self):
        return self.__package_num


class OTADevice(object):
    def __init__(self, filename):
        # self._device_address = {
        #     '/dev/ttyHSL1': {
        #         'para': [19200, 'N', 8, 1],
        #         'devs': [
        #           {
        #               'slave': 1,
        #               'mode': [03,1000], #功能码，地址
        #               'power': [03,1000]
        #           }
        #         ]
        #     }
        # }
        self._device_para = dict()
        self._initial_from_config(filename)

    def get_coms(self):
        return self._device_para.keys()

    def get_sub_devices(self, com=''):
        devs = list()
        for dev in self._device_para[com].get('devs', list()):
            devs.append((dev.get('slave'), dev.get('mode', list()), dev.get('power', list())))
        return devs

    def _initial_from_config(self, filename):
        Log.info('解析配置文件：%s' % filename)
        tree = ET.parse(filename)
        for model in tree.getroot().findall('./model[@d="T.IDrv.Task:libTModbus"]'):
            config = model.get('config')
            if config is None:
                continue
            Log.debug('获取通讯串口配置参数：' + str(config))
            mode, port, baud, parity, bytesize, stopzise = str(config).split(';')[:6]
            para = list()
            devs = list()
            if str(mode).upper() in ['RTU', 'ASCII']:
                para.extend([int(baud), str(parity).upper()[0], int(bytesize), int(stopzise)])
                for driver in model.findall('./device/driver[@n="T.IDrv.Task:libTModbus"]'):
                    driver_config = driver.get('config')
                    if driver_config is None:
                        continue
                    Log.debug('获取主板从站地址：' + str(driver_config))
                    item = driver.find('./commDataItems/dataItem[@n="power"]')
                    if item is None:
                        continue
                    item_config = item.get('config')
                    if item_config is None:
                        continue
                    Log.debug('获取电源（power）配置：' + str(item_config))
                    item_config_l = str(item_config).split(';')
                    power_func_code = int(item_config_l[0])
                    reg_address_str = item_config_l[1].upper()
                    if reg_address_str.isdigit():
                        power_reg_address = int(reg_address_str)
                    elif reg_address_str.startswith('0X'):
                        power_reg_address = int(reg_address_str, 16)
                    else:
                        continue
                    item = driver.find('./commDataItems/dataItem[@n="mode"]')
                    if item is None:
                        continue
                    item_config = item.get('config')
                    if item_config is None:
                        continue
                    Log.debug('获取模式（mode）配置：' + str(item_config))
                    item_config_l = str(item_config).split(';')
                    mode_func_code = int(item_config_l[0])
                    reg_address_str = item_config_l[1].upper()
                    if reg_address_str.isdigit():
                        mode_reg_address = int(reg_address_str)
                    elif reg_address_str.startswith('0X'):
                        mode_reg_address = int(reg_address_str, 16)
                    else:
                        continue
                    devs.append(
                        {
                            'slave': int(str(driver_config).split(';')[0]),
                            'mode': [mode_func_code, mode_reg_address],
                            'power': [power_func_code, power_reg_address],
                        }
                    )

            if len(para) and len(devs):
                self._device_para[port] = {
                    'para': para,
                    'devs': devs
                }
                Log.info("获取主板配置参数:" + json.dumps(self._device_para))

        if len(self._device_para.keys()) <= 0:
            raise OTAError('请配置Modbus任务驱动，并且配置power和mode数据项 ！')


class OTAProcess(object):
    # 0：无模式
    UP_MODE_NONE = 0
    # 1：数据接收模式
    UP_MODE_DATA_RECV = 1
    # 2：内板程序更新模式
    UP_MODE_IN_PROGRAM = 2
    # 3：内板EEPROM更新模式
    UP_MODE_IN_EEPROM = 3
    # 4：外机1程序更新模式
    UP_MODE_OUT1_PROGRAM = 4
    # 5：外机1 EEPROM更新模式
    UP_MODE_OUT1_EEPROM = 5
    # 6：外机2程序更新模式
    UP_MODE_OUT2_PROGRAM = 6
    # 7：外机2 EEPROM更新模式
    UP_MODE_OUT2_EEPROM = 7

    def __init__(self, slave=0, ota_mode='', dev_mode=None, dev_power=None, master=None):
        isinstance(master, modbus_rtu.RtuMaster)
        self.master = master
        self.slave = slave
        self.ota_mode = self.check_ota_mode(ota_mode)
        self.dev_mode = dev_mode  # [03, 1000]
        self.dev_mode_val = -1
        self.dev_power = dev_power  # [03, 1000]
        self.dev_power_val = -1
        self.file = OTAFile(os.path.join(WORK_DIR, 'image'))

        self._run = True
        self._update_cmd_status_code = 0
        self._update_error_protect_code = 0

        self._ota_code = UpdateCode.FAILED
        self._ota_msg = ''

    def get_result(self):
        return OTAResult(code=self._ota_code, info=self._ota_msg)

    @staticmethod
    def check_ota_mode(mode_name=''):
        if mode_name == 'MODE2':
            return OTAProcess.UP_MODE_IN_PROGRAM
        elif mode_name == 'MODE3':
            return OTAProcess.UP_MODE_IN_EEPROM
        elif mode_name == 'MODE4':
            return OTAProcess.UP_MODE_OUT1_PROGRAM
        elif mode_name == 'MODE5':
            return OTAProcess.UP_MODE_OUT1_EEPROM
        elif mode_name == 'MODE6':
            return OTAProcess.UP_MODE_OUT2_PROGRAM
        elif mode_name == 'MODE7':
            return OTAProcess.UP_MODE_OUT2_EEPROM
        else:
            return -1

    def get_ota_mode_name(self):
        name = '（%s）' % str(self.ota_mode)
        if self.ota_mode == 2:
            name += '内板程序更新模式'
        elif self.ota_mode == 3:
            name += '内板EEPROM更新模式'
        elif self.ota_mode == 4:
            name += '外机1程序更新模式'
        elif self.ota_mode == 5:
            name += '外机1 EEPROM更新模式'
        elif self.ota_mode == 6:
            name += '外机2程序更新模式'
        elif self.ota_mode == 7:
            name += '外机2 EEPROM更新模式'
        return name

    def _modbus_read_reg(self, slave, func_code, address, length):
        try_num = 3
        ts = 0
        timeout_s = 3
        self.master.set_timeout(1)
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    reply = self.master.execute(slave, func_code, address, length)
                    if not reply:
                        raise OTAError(info='not reply')
                    return reply
            except Exception as ept:
                try_num -= 1
                if try_num <= 0:
                    raise OTAError(info='Modbus查询寄存器失败, ' + repr(ept))
            finally:
                time.sleep(0.01)

    def _modbus_get_mode_and_power_status(self):
        if not self.dev_mode:
            raise OTAError(info='未找到主板模式配置！')
        if not self.dev_power:
            raise OTAError(info='未找到主板电源配置！')

        Log.info('查询当前设备状态')
        reply = self._modbus_read_reg(self.slave, self.dev_power[0], self.dev_power[1], 1)
        if reply:
            self.dev_power_val = reply[0]
        if self.dev_mode[0] == self.dev_power[0] \
                and self.dev_mode[1] == self.dev_power[1]:  # 电源和模式地址相同
            self.dev_mode_val = self.dev_power_val
        else:
            reply = self._modbus_read_reg(self.slave, self.dev_mode[0], self.dev_mode[1], 1)
            if reply:
                self.dev_mode_val = reply[0]
        Log.info('当前设备状态，电源：%d，模式：%d' % (self.dev_power_val, self.dev_mode_val))

    def _check_self(self):
        if not (0 < self.slave <= 254):
            raise OTAError(info='从站ID（%d）范围应在(0< id <= 254)！' % self.slave)
        if self.ota_mode == -1:
            raise OTAError(info='升级模式（%d）范围应在(2 <= id <= 7)！' % self.ota_mode)
        if not self.file.check_self():
            raise OTAError(info='未找到升级文件！')
        self._modbus_get_mode_and_power_status()

    def _do_transport_unlock_device(self):
        try_num = 3
        ts = 0
        timeout_s = 3
        self.master.set_timeout(3)
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    reply = self.master.execute(self.slave, 0X06, 10050 - 1, output_value=0X0001)
                    if not reply:
                        raise OTAError(info='not reply')
                    else:
                        time.sleep(1)
                        return
            except Exception as ept:
                try_num -= 1
                if try_num <= 0:
                    raise OTAError(info='解锁机器失败，无数据响应，' + repr(ept))
            finally:
                time.sleep(0.01)

    def _do_transport_start_sendfile_mode(self):
        try_num = 3
        ts = 0
        timeout_s = 3
        self.master.set_timeout(1.5)
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    reply = self.master.execute(self.slave, 0X10, 10050 - 1,
                                                output_value=[0X0001,
                                                              0X0101,
                                                              0x5AA5,
                                                              self.file.filetype,
                                                              self.file.package_num
                                                              ])
                    if not reply:
                        raise OTAError(info='not reply')
                    else:
                        time.sleep(3)
                        return
            except Exception as ept:
                try_num -= 1
                if try_num <= 0:
                    raise OTAError(info='设置升级模式失败，无数据响应,' + repr(ept))
            finally:
                time.sleep(0.01)

    def _modbus_query_update_status(self):
        try:
            self._update_error_protect_code = -1
            self._update_cmd_status_code = -1
            self.master.set_timeout(1.5)
            reply = self.master.execute(self.slave, 0X03, 10060 - 1, 2)
            if reply:
                self._update_cmd_status_code = reply[0]
                self._update_error_protect_code = reply[1]
                Log.info('远程升级状态: %d，远程升级故障码：%d' % (self._update_cmd_status_code, self._update_cmd_status_code))
                return True
        except Exception as ept:
            Log.warn('Modbus查询失败')
        return False

    def _do_transport_query_update_status(self):
        try_num = 3
        ts = 0
        timeout_s = 3
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    if self._modbus_query_update_status():
                        if self._update_cmd_status_code == 3:
                            Log.info('远程升级进入传输过程')
                            return
                        else:
                            if self._update_cmd_status_code == 0:
                                self._ota_msg = '远程升级故障码(%s)' % str(self._update_error_protect_code)
                            raise OTAError(info='远程升级状态(%s) != 3，远程升级故障码：%s' % (str(self._update_cmd_status_code), str(self._update_error_protect_code)))
                    else:
                        raise OTAError(info='not reply')
            except Exception as ept:
                try_num -= 1
                if try_num <= 0:
                    raise OTAError(info='查询远程升级状态和故障码状态失败，' + repr(ept))
            finally:
                time.sleep(0.01)

    def _do_transport_send_file(self):
        reply_failed_cnt = 0
        reply_failed_max_num = 3
        self.master.set_timeout(3)
        head_str = ''
        transport_percent_scale = 100.0 / self.file.package_num
        transport_process = 0
        while self._run:
            eof, index, length, data = self.file.get_next_package()
            try:
                request = struct.pack('BB', self.slave, 0X15)
                request += struct.pack('<H', index)
                if eof:
                    request += struct.pack('B', 01)
                else:
                    request += struct.pack('B', 00)
                request += struct.pack('B', 00)
                request += struct.pack('B', length)
                request += data
                request += struct.pack('>H', utils.calculate_crc(request))
                expected_length = 9
                head_str = '(%d/%d)' % (index, self.file.package_num)
                Log.debug(utils.get_log_buffer(head_str + " -> ", request))
                self.master._send(request)
                reply = self.master._recv(expected_length)
                Log.debug(utils.get_log_buffer(head_str + " <- ", reply))
                process = int(index * transport_percent_scale)
                if not (process % 25) and process != transport_process:
                    transport_process = process
                    Log.info('文件传输进度：{}%'.format(transport_process))
                if reply \
                        and len(reply) == expected_length \
                        and struct.unpack('B', reply[0])[0] == self.slave \
                        and struct.unpack('B', reply[1])[0] == 0X15 \
                        and utils.calculate_crc(reply[:7]) == struct.unpack('>H', reply[-2:])[0]:  # crc
                    reply_failed_cnt = 0
                    recv_index = struct.unpack('>H', reply[2:4])[0]
                    end_ok = struct.unpack('B', reply[4])[0]
                    burn_ok = struct.unpack('B', reply[5])[0]
                    if not eof:
                        if recv_index != index:
                            if burn_ok == 0:
                                self.file.set_next_package_index(recv_index)
                            else:
                                self.file.set_next_package_index(recv_index + 1)
                        else:
                            if burn_ok == 0:
                                self.file.set_next_package_index(index)
                    else:
                        if recv_index == 0X0000 and burn_ok == 0X01 and end_ok == 0X01:
                            Log.info(head_str + '结束包已烧录，传输升级文件成功')
                            return
                        else:
                            self.file.set_next_package_index(index)
                else:
                    reply_failed_cnt += 1
                    if reply_failed_cnt >= reply_failed_max_num:
                        raise OTAError(info='not reply')
                    else:
                        self.file.set_next_package_index(index)
                        time.sleep(3)
                        continue
            except OTAError as ept:
                self._modbus_query_update_status()
                if eof:
                    if self._update_cmd_status_code == 0 and self._update_error_protect_code == 0:
                        Log.info(head_str + '查询升级状态，传输升级文件成功')
                        return
                    else:
                        raise OTAError(info='传输[结束包数据]无响应，查询传输结状态显示失败status_code（%s）, error_code（%s）!'
                                            % (str(self._update_cmd_status_code), str(self._update_error_protect_code)))
                else:
                    raise OTAError(info='传输[非结束包]数据无响应，' + repr(ept))
            except Exception as ept:
                raise OTAError(info=repr(ept))

    def _do_update_set_otastarting_mode(self):
        try_num = 3
        ts = 0
        timeout_s = 3
        self.master.set_timeout(1.5)
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    reply = self.master.execute(self.slave, 0X10, 10050 - 1,
                                                output_value=[0X0001,
                                                              0X0100 | (self.ota_mode & 0XFF),
                                                              0x5AA5,
                                                              self.file.filetype,
                                                              self.file.package_num
                                                              ])
                    if not reply:
                        raise OTAError(info='not reply')
                    else:
                        Log.info('设置升级开始指令成功，10秒后查询升级结果')
                        time.sleep(10)
                        return
            except Exception as ept:
                try_num -= 1
                if try_num <= 0:
                    raise OTAError(info='设置升级开始指令失败，' + repr(ept))
            finally:
                time.sleep(0.01)

    def _do_update_query_update_result(self):
        try_num = 600
        ts = 0
        timeout_s = 1
        self.master.set_timeout(1)
        while self._run:
            try:
                if (time.time() - ts) > timeout_s:
                    ts = time.time()
                    if self._modbus_query_update_status():
                        if self._update_cmd_status_code == 0:
                            if self._update_error_protect_code == 0:
                                self._ota_code = UpdateCode.SUCCESS
                                Log.info('查询升级状态，升级成功，远程升级状态（%s）' % self._update_error_protect_code)
                                return  # 升级成功
                            elif self._update_error_protect_code in [9, 11]:
                                raise OTAError(info='查询升级状态，升级失败，远程升级状态（%s）' % self._update_error_protect_code)
                            else:
                                try_num -= 1
                                if try_num <= 0:
                                    raise OTAError(info='远程升级故障码(%d)' % self._update_error_protect_code)
                                else:
                                    pass    # 继续查询
                        else:
                            try_num -= 1
                            if try_num <= 0:
                                raise OTAError(info='远程升级状态(%d)' % self._update_cmd_status_code)
                            else:
                                pass  # 继续查询
                    else:
                        try_num -= 1
                        if try_num <= 0:
                            raise OTAError(info='查询升级状态失败，无数据响应')
                        else:
                            pass  # 继续查询
            except Exception as ept:
                raise OTAError(info=repr(ept))
            finally:
                time.sleep(0.01)

    def _do_suspend_device(self):
        try:
            if self.dev_power_val > 0:
                Log.info('关闭机器电源，地址：%d' % self.dev_power[1])
                self.master.execute(self.slave, 0X06, self.dev_power[1], output_value=0)
                Log.info('等待机器关机，等待60秒')
                time.sleep(60)
            else:
                Log.info('机器已关机')
        except Exception as ept:
            Log.warn('中断机器运行失败，' + repr(ept))

    def _do_restore_device(self):
        try:
            if self.dev_power_val > 0:
                self.master.set_timeout(1)
                if self.dev_mode[0] == self.dev_power[0] \
                        and self.dev_mode[1] == self.dev_power[1]:  # 电源和模式地址相同
                    Log.info('打开机器电源，同时设置运行模式，地址：%d，模式：%d' % (self.dev_mode[1], self.dev_mode_val))
                    self.master.execute(self.slave, 0X06, self.dev_mode[1], output_value=self.dev_mode_val)  # 写入模式的值
                else:
                    Log.info('打开机器电源：（%d, %d）' % (self.dev_power[1], self.dev_power_val))
                    self.master.execute(self.slave, 0X06, self.dev_power[1],
                                        output_value=self.dev_power_val)  # 写入电源的值
            else:
                Log.info('机器保持关机状态')
        except Exception as ept:
            Log.warn('恢复机器运行失败，' + repr(ept))

    def _ota_process_do(self):
        self._do_suspend_device()
        try:
            self._do_transport_unlock_device()
            self._do_transport_start_sendfile_mode()
            self._do_transport_query_update_status()
            self._do_transport_send_file()
            self._do_update_set_otastarting_mode()
            self._do_update_query_update_result()
        except OTAError as ept:
            self._ota_code = ept.code
            self._ota_msg = ept.message
        except Exception as ept:
            self._ota_code = UpdateCode.EXCEPTION
            self._ota_msg = repr(ept)
            Log.warn('升级失败，' + repr(ept))
        finally:
            self._do_restore_device()

    def run(self):
        s_ts = time.time()
        try:
            Log.info('开始主板（%d）升级，升级模式：%s' % (self.slave, self.get_ota_mode_name()))
            self._check_self()
            self._ota_process_do()
        except Exception as ept:
            self._ota_code = UpdateCode.EXCEPTION
            self._ota_msg = repr(ept)
            Log.warn('升级失败，' + repr(ept))
        finally:
            Log.info('退出主板（%d）升级，升级模式：%s，耗时：%.3f 秒' % (self.slave, self.get_ota_mode_name(), time.time() - s_ts))


class AmitimeOta(DeviceOTABase):

    def __init__(self, root_path, tag1, tag2, tag3, *args, **kwargs):
        super(AmitimeOta, self).__init__(root_path, tag1, tag2, tag3)
        self._ota_devices = OTADevice(filename=AgentBasic.get_module_config())
        self._ota_thread = list()

        self._ota_result = list()

    def __open_modbus_master(self, port):
        try:
            master = modbus_rtu.RtuMaster(serial.Serial(
                port=port,
                baudrate=19200,
                parity='E',
                bytesize=8,
                stopbits=1
            ))
            master.set_timeout(3)
            master.set_verbose(True)
            return master
        except Exception as ept:
            pass
        return None

    def _ota_process(self, com):
        master = self.__open_modbus_master(com)
        for slave, mode, power in self._ota_devices.get_sub_devices(com):
            process = OTAProcess(slave=slave,
                                 ota_mode=self.tag3,
                                 dev_mode=mode,
                                 dev_power=power,
                                 master=master)
            process.run()
            self._ota_result.append(process.get_result())

        if master:
            master._serial.close()
            master.close()

    def _run(self):
        for com in self._ota_devices.get_coms():
            thread = threading.Thread(target=self._ota_process, args=(com,), name=com)
            thread.setDaemon(True)
            self._ota_thread.append(thread)

        for thread in self._ota_thread:
            thread.start()

        for thread in self._ota_thread:
            thread.join()

    def run(self):
        initial_logger(
            filepath=Log.get_log_file(),
            enable_debug=ENABLE_DEBUG,
            enable_stdout=ENABLE_DEBUG
        )
        ts = time.time()
        try:
            Log.info('>> 进入程序 %s %s' % (PROGRAM_NAME, PROGRAM_VERSION))
            Log.info('>> 启动指令 %s' % str(sys.argv))
            AgentBasic.exit()

            self._run()

            for result in self._ota_result:
                if result.code == UpdateCode.SUCCESS:
                    self.result_code = UpdateCode.SUCCESS
                    break
        except OTAError as e:
            self.result_code = e.code
            self.result_message = e.message
            Log.warn('告警，' + str(e))
        except Exception as e:
            self.result_code = UpdateCode.EXCEPTION
            self.result_message = repr(e)
            Log.warn('告警，' + repr(e))
        finally:
            AgentBasic.entry()
            Log.info('>> 退出程序，状态：%s, 耗时: %.3f 秒' % (str(self.result_code), time.time() - ts))

